package org.w3c.rdf.implementation.model;

import java.io.*;
import java.util.*;

/**
 * LongTable collision list.
 */
class LongTableEntry {
  long hash;
  //    Object key; key.hashCode() = hash
  Object value;
  LongTableEntry next;

  protected Object clone() {
    LongTableEntry entry = new LongTableEntry();
    entry.hash = hash;
    //    entry.key = key;
    entry.value = value;
    entry.next = (next != null) ? (LongTableEntry)next.clone() : null;
    return entry;
  }
}

/**
 * A hashtable specialized for storing keys of type <code>long</code>. Supports duplicates.
 */

public class LongTable implements Cloneable, java.io.Serializable {
    /**
     * The hash table data.
     */
    private transient LongTableEntry table[];

    /**
     * The total number of entries in the hash table.
     */
    private transient int count;

    /**
     * Rehashes the table when count exceeds this threshold.
     */
    private int threshold;

    /**
     * The load factor for the hashtable.
     */
    private float loadFactor;

  /**
   * true if multiple hash values allowed
   */
    private boolean multivalues;

    /**
     * Constructs a new, empty hashtable with the specified initial 
     * capacity and the specified load factor. 
     *
     * @param      initialCapacity   the initial capacity of the hashtable.
     * @param      loadFactor        a number between 0.0 and 1.0.
     * @exception  IllegalArgumentException  if the initial capacity is less
     *               than or equal to zero, or if the load factor is less than
     *               or equal to zero.
     * @since      JDK1.0
     */
    public LongTable(int initialCapacity, float loadFactor, boolean multivalues) {
	if ((initialCapacity <= 0) || (loadFactor <= 0.0)) {
	    throw new IllegalArgumentException();
	}
	this.loadFactor = loadFactor;
	table = new LongTableEntry[initialCapacity];
	threshold = (int)(initialCapacity * loadFactor);
	this.multivalues = multivalues;
    }

    /**
     * Constructs a new, empty hashtable with the specified initial capacity
     * and default load factor.
     *
     * @param   initialCapacity   the initial capacity of the hashtable.
     * @since   JDK1.0
     */
    public LongTable(int initialCapacity) {
	this(initialCapacity, 0.75f, false);
    }

    /**
     * Constructs a new, empty hashtable with a default capacity and load
     * factor. 
     *
     * @since   JDK1.0
     */
    public LongTable() {
	this(101, 0.75f, false);
    }

    public LongTable(boolean multivalues) {
	this(101, 0.75f, multivalues);
    }

    /**
     * Returns the number of keys in this hashtable.
     *
     * @return  the number of keys in this hashtable.
     * @since   JDK1.0
     */
    public int size() {
	return count;
    }

    /**
     * Tests if this hashtable maps no keys to values.
     *
     * @return  <code>true</code> if this hashtable maps no keys to values;
     *          <code>false</code> otherwise.
     * @since   JDK1.0
     */
    public boolean isEmpty() {
	return count == 0;
    }

    /**
     * Returns an enumeration of the keys in this hashtable.
     *
     * @return  an enumeration of the keys in this hashtable.
     * @see     java.util.Enumeration
     * @see     java.util.LongTable#elements()
     * @since   JDK1.0
     */
    public synchronized LongEnumeration keys() {
	return new KeysEnumerator(table);
    }

    /**
     * Returns an enumeration of the values in this hashtable.
     * Use the Enumeration methods on the returned object to fetch the elements
     * sequentially.
     *
     * @return  an enumeration of the values in this hashtable.
     * @see     java.util.Enumeration
     * @see     java.util.LongTable#keys()
     * @since   JDK1.0
     */
    public synchronized Enumeration elements() {
	return new ValuesEnumerator(table);
    }

    public synchronized Enumeration multiget(long key) {
      return new MultiEnumerator(table, key);
    }

  public synchronized int multigetsize(long key) {
    
    int index = (int)((key & 0x7FFFFFFFFFFFFFFFl) % table.length);
    LongTableEntry entry = table[index];
    int counter = 0;
    
    while(entry != null) {
      if(entry.hash == key)
	counter++;
      entry = entry.next;
    }
    return counter;
  }

    /**
     * Tests if some key maps into the specified value in this hashtable.
     * This operation is more expensive than the <code>containsKey</code>
     * method.
     *
     * @param      value   a value to search for.
     * @return     <code>true</code> if some key maps to the
     *             <code>value</code> argument in this hashtable;
     *             <code>false</code> otherwise.
     * @exception  NullPointerException  if the value is <code>null</code>.
     * @see        java.util.LongTable#containsKey(java.lang.Object)
     * @since      JDK1.0
     */
    public synchronized boolean contains(Object value) {
	if (value == null) {
	    throw new NullPointerException();
	}

	LongTableEntry tab[] = table;
	for (int i = tab.length ; i-- > 0 ;) {
	    for (LongTableEntry e = tab[i] ; e != null ; e = e.next) {
		if (e.value.equals(value)) {
		    return true;
		}
	    }
	}
	return false;
    }

    /**
     * Tests if the specified object is a key in this hashtable.
     * 
     * @param   key   possible key.
     * @return  <code>true</code> if the specified object is a key in this
     *          hashtable; <code>false</code> otherwise.
     * @see     java.util.LongTable#contains(java.lang.Object)
     * @since   JDK1.0
     */
    public synchronized boolean containsKey(long hash) {
	LongTableEntry tab[] = table;
	int index = (int)((hash & 0x7FFFFFFFFFFFFFFFl) % tab.length);
	for (LongTableEntry e = tab[index] ; e != null ; e = e.next) {
	    if (e.hash == hash) {
		return true;
	    }
	}
	return false;
    }

    public synchronized boolean contains(long hash, Object value) {
	LongTableEntry tab[] = table;
	int index = (int)((hash & 0x7FFFFFFFFFFFFFFFl) % tab.length);
	for (LongTableEntry e = tab[index] ; e != null ; e = e.next) {
	    if (e.hash == hash && e.value.equals(value)) {
		return true;
	    }
	}
	return false;
    }

    /**
     * Returns the value to which the specified key is mapped in this hashtable.
     *
     * @param   key   a key in the hashtable.
     * @return  the value to which the key is mapped in this hashtable;
     *          <code>null</code> if the key is not mapped to any value in
     *          this hashtable.
     * @see     java.util.LongTable#put(java.lang.Object, java.lang.Object)
     * @since   JDK1.0
     */
    public synchronized Object get(long hash) {
	LongTableEntry tab[] = table;
	int index = (int)((hash & 0x7FFFFFFFFFFFFFFFl) % tab.length);
	for (LongTableEntry e = tab[index] ; e != null ; e = e.next) {
	  if (e.hash == hash) {
	    return e.value;
	  }
	}
	return null;
    }

    /**
     * Rehashes the contents of the hashtable into a hashtable with a 
     * larger capacity. This method is called automatically when the 
     * number of keys in the hashtable exceeds this hashtable's capacity 
     * and load factor. 
     *
     * @since   JDK1.0
     */
    protected void rehash() {
	int oldCapacity = table.length;
	LongTableEntry oldTable[] = table;

	int newCapacity = oldCapacity * 2 + 1;
	LongTableEntry newTable[] = new LongTableEntry[newCapacity];

	threshold = (int)(newCapacity * loadFactor);
	table = newTable;

	//System.out.println("rehash old=" + oldCapacity + ", new=" + newCapacity + ", thresh=" + threshold + ", count=" + count);

	for (int i = oldCapacity ; i-- > 0 ;) {
	    for (LongTableEntry old = oldTable[i] ; old != null ; ) {
		LongTableEntry e = old;
		old = old.next;

		int index = (int)((e.hash & 0x7FFFFFFFFFFFFFFFl) % newCapacity);
		e.next = newTable[index];
		newTable[index] = e;
	    }
	}
    }

    /**
     * Maps the specified <code>key</code> to the specified 
     * <code>value</code> in this hashtable. Neither the key nor the 
     * value can be <code>null</code>. 
     * <p>
     * The value can be retrieved by calling the <code>get</code> method 
     * with a key that is equal to the original key. 
     *
     * @param      key     the hashtable key.
     * @param      value   the value.
     * @return     the previous value of the specified key in this hashtable,
     *             or <code>null</code> if it did not have one.
     * @exception  NullPointerException  if the key or value is
     *               <code>null</code>.
     * @see     java.lang.Object#equals(java.lang.Object)
     * @see     java.util.LongTable#get(java.lang.Object)
     * @since   JDK1.0
     */
    public synchronized Object put(long hash, Object value) {
	// Make sure the value is not null
	if (value == null) {
	    throw new NullPointerException();
	}

	// Makes sure the key is not already in the hashtable.
	LongTableEntry tab[] = table;
	int index = (int)((hash & 0x7FFFFFFFFFFFFFFFl) % tab.length);
	for (LongTableEntry e = tab[index] ; e != null ; e = e.next) {
	    if ((e.hash == hash) && (!multivalues || e.value.equals(value))) {
		Object old = e.value;
		e.value = value;
		return old;
	    }
	}

	if (count >= threshold) {
	    // Rehash the table if the threshold is exceeded
	    rehash();
	    return put(hash, value);
	} 

	// Creates the new entry.
	LongTableEntry e = new LongTableEntry();
	e.hash = hash;
	e.value = value;
	e.next = tab[index];
	tab[index] = e;
	count++;
	return null;
    }

    /**
     * Removes the key (and its corresponding value) from this 
     * hashtable. This method does nothing if the key is not in the hashtable.
     *
     * @param   key   the key that needs to be removed.
     * @return  the value to which the key had been mapped in this hashtable,
     *          or <code>null</code> if the key did not have a mapping.
     * @since   JDK1.0
     */
    public synchronized Object remove(long hash) {

      return remove(hash, null);
    }

    public synchronized Object remove(long hash, Object value) {
	LongTableEntry tab[] = table;
	int index = (int)((hash & 0x7FFFFFFFFFFFFFFFl) % tab.length);
	for (LongTableEntry e = tab[index], prev = null ; e != null ; prev = e, e = e.next) {
	    if ((e.hash == hash) && (value == null || e.value.equals(value))) {
		if (prev != null) {
		    prev.next = e.next;
		} else {
		    tab[index] = e.next;
		}
		count--;
		return e.value;
	    }
	}
	return null;
    }

    /**
     * Clears this hashtable so that it contains no keys. 
     *
     * @since   JDK1.0
     */
    public synchronized void clear() {
	LongTableEntry tab[] = table;
	for (int index = tab.length; --index >= 0; )
	    tab[index] = null;
	count = 0;
    }

    /**
     * Creates a shallow copy of this hashtable. The keys and values 
     * themselves are not cloned. 
     * This is a relatively expensive operation.
     *
     * @return  a clone of the hashtable.
     * @since   JDK1.0
     */
    public synchronized Object clone() {
	try { 
	    LongTable t = (LongTable)super.clone();
	    t.table = new LongTableEntry[table.length];
	    for (int i = table.length ; i-- > 0 ; ) {
		t.table[i] = (table[i] != null) 
		    ? (LongTableEntry)table[i].clone() : null;
	    }
	    return t;
	} catch (CloneNotSupportedException e) { 
	    // this shouldn't happen, since we are Cloneable
	    throw new InternalError();
	}
    }

    /**
     * Returns a rather long string representation of this hashtable.
     *
     * @return  a string representation of this hashtable.
     * @since   JDK1.0
     */
    public synchronized String toString() {
	int max = size() - 1;
	StringBuffer buf = new StringBuffer();
	LongEnumeration k = keys();
	Enumeration e = elements();
	buf.append("{");

	for (int i = 0; i <= max; i++) {
	    String s1 = String.valueOf(k.nextElement());
	    String s2 = e.nextElement().toString();
	    buf.append(s1 + "=" + s2);
	    if (i < max) {
		buf.append(", ");
	    }
	}
	buf.append("}");
	return buf.toString();
    }

    /**
     * WriteObject is called to save the state of the hashtable to a stream.
     * Only the keys and values are serialized since the hash values may be
     * different when the contents are restored.
     * iterate over the contents and write out the keys and values.
     */
    private synchronized void writeObject(java.io.ObjectOutputStream s)
        throws IOException
    {
	// Write out the length, threshold, loadfactor
	s.defaultWriteObject();

	// Write out length, count of elements and then the key/value objects
	s.writeInt(table.length);
	s.writeInt(count);
	for (int index = table.length-1; index >= 0; index--) {
	    LongTableEntry entry = table[index];

	    while (entry != null) {
		s.writeLong(entry.hash);
		s.writeObject(entry.value);
		entry = entry.next;
	    }
	}
    }

    /**
     * readObject is called to restore the state of the hashtable from
     * a stream.  Only the keys and values are serialized since the
     * hash values may be different when the contents are restored.
     * Read count elements and insert into the hashtable. 
     */
    private synchronized void readObject(java.io.ObjectInputStream s)
         throws IOException, ClassNotFoundException
    {
	// Read in the length, threshold, and loadfactor
	s.defaultReadObject();

	// Read the original length of the array and number of elements
	int origlength = s.readInt();
	int elements = s.readInt();

	// Compute new size with a bit of room 5% to grow but
	// No larger than the original size.  Make the length
	// odd if it's large enough, this helps distribute the entries.
	// Guard against the length ending up zero, that's not valid.
	int length = (int)(elements * loadFactor) + (elements / 20) + 3;
	if (length > elements && (length & 1) == 0)
	    length--;
	if (origlength > 0 && length > origlength)
	    length = origlength;

	table = new LongTableEntry[length];
	count = 0;

	// Read the number of elements and then all the key/value objects
	for (; elements > 0; elements--) {
	    long hash = s.readLong();
	    Object value = s.readObject();
	    put(hash, value);
	}
    }

  public static void main(String[] args) {

    LongTable numbers = new LongTable(true);
    numbers.put(1, "one");
    numbers.put(2, "two");
    numbers.put(3, "three");
    numbers.put(3, "three2");
    numbers.put(3, "three3");
    numbers.put(3, "three");
    numbers.put(4, "four");

    System.out.println("3: " + numbers.get(3));

    System.out.println("multigetsize(3): " + numbers.multigetsize(3));
    
    for(Enumeration en = numbers.multiget(3); en.hasMoreElements();)
      System.out.println("value: " + en.nextElement());
  }

interface LongEnumeration {

  public boolean hasMoreElements();
  public long nextElement();
}

/**
 * Enumerates LongTableEntries
 */
class LongTableEnumerator {

    int index;
    LongTableEntry table[];
    LongTableEntry entry;

    LongTableEnumerator(LongTableEntry table[]) {
	this.table = table;
	this.index = table.length;
    }
	
    public boolean hasMoreElements() {
	if (entry != null) {
	    return true;
	}
	while (index-- > 0) {
	    if ((entry = table[index]) != null) {
		return true;
	    }
	}
	return false;
    }

    protected LongTableEntry getNextElement() {
	if (entry == null) {
	    while ((index-- > 0) && ((entry = table[index]) == null));
	}
	if (entry != null) {
	    LongTableEntry e = entry;
	    entry = e.next;
	    return e;
	}
	throw new NoSuchElementException("LongTableEnumerator");
    }
}

class ValuesEnumerator extends LongTableEnumerator implements Enumeration {

    ValuesEnumerator(LongTableEntry table[]) {
      super(table);
    }
	
    public Object nextElement() {
      return getNextElement().value;
    }
}

class KeysEnumerator extends LongTableEnumerator implements LongEnumeration {

    KeysEnumerator(LongTableEntry table[]) {
      super(table);
    }
	
    public long nextElement() {
      return getNextElement().hash;
    }
}


class MultiEnumerator implements Enumeration {

    LongTableEntry entry;
    long hash;

    MultiEnumerator(LongTableEntry table[], long hash) {
        this.hash = hash;
	int index = (int)((hash & 0x7FFFFFFFFFFFFFFFl) % table.length);
	entry = table[index];
	// position
	while(entry != null && entry.hash != hash)
	  entry = entry.next;
    }
	
    public boolean hasMoreElements() {
      return entry != null;
    }

    public Object nextElement() {
      if (entry != null) {
	LongTableEntry e = entry;
	do {
	  entry = entry.next;
	} while (entry != null && hash != entry.hash);
	return e.value;
      }
      throw new NoSuchElementException("MultiEnumerator");
    }
}

}
